#!/usr/bin/perl -w

######################################################################
#
# Copyright (C) 2010 Sean Adams and Richard Perlman
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
######################################################################

#unshift(@INC, '.', '..');

use strict;

use Term::ANSIColor;
use Device::SerialPort qw( :PARAM :STAT 0.07 );
use Time::HiRes qw(time);
use List::Util qw(shuffle);
use Data::Dumper;
require "getopts.pl";

our ($opt_d, $opt_h, $opt_i, $opt_m, $opt_o);
&Getopts('d:h:i:m:o:');

if (! defined $opt_d) {
	print STDERR "$0 usage:\n    -d Device file name (required)\n    -h hops\n    -i iterations\n    -m interface\n    -o log file\n\n";
	exit;
}

my $DEVICE_LIST_FILENAME = $opt_d;
my ($ITERATIONS, $INITIAL_HOPS, $INTERFACE, $LOGDIR, $RPTFIL);

if (defined $opt_h) {
	$INITIAL_HOPS = $opt_h;
} else {
	$INITIAL_HOPS = 1;
}

if (defined $opt_i) {
	$ITERATIONS = $opt_i;
} else {
	$ITERATIONS = 100;
}

if (defined $opt_m) {
	$INTERFACE = $opt_m;
}  else {
	$INTERFACE = "notset";
}

if (defined $opt_o) {
	$LOGDIR = `/usr/bin/dirname $opt_o`;
	chomp($LOGDIR);
	$LOGDIR = $LOGDIR  . "/iHealth_logs";
	$RPTFIL = $opt_o;
}  else {
	$LOGDIR = "/var/tmp/iHealth_logs";
	$RPTFIL = "/var/tmp/iHealth-Report.txt";
}

sub array_of_bytes_to_text {

	sprintf "".("%02X " x scalar(@_))."\n", @_;
}

sub packed_to_text {
	my $s = shift;

	&array_of_bytes_to_text(unpack "C" x length($s), $s);
}

sub usb_connect {

	## rdp
	my ($PORTNAME) = @_;
	
	if ($PORTNAME =~ "notset") {
		my $eachdev;
		my @usbdevices = ();
	
	
		opendir DEVDIR, "/dev";
		while ($eachdev = readdir(DEVDIR)) {
			if ($eachdev=~/cu.usbserial/) {
				push @usbdevices, "/dev/$eachdev";
			}
		}
	
		my $numdevs = (@usbdevices);
	
		if ($numdevs > 1) {
			print STDOUT "found $numdevs interface(s): ", join(', ', @usbdevices), "\n";
			print REPORT "found $numdevs interface(s): ", join(', ', @usbdevices), "\n";
		} else {
			if (!$numdevs) {
				print STDOUT "No interfaces found.\n";
				print REPORT "No interfaces found.\n";
			return;
			}
		}
	
		$PORTNAME = pop(@usbdevices);
	} else {
		$PORTNAME = "/dev/" . $PORTNAME;
	}
	## rdp
	print STDOUT "using interface $PORTNAME\n";
	print REPORT "using interface $PORTNAME\n";
	
	my $PortObj = new Device::SerialPort ($PORTNAME, 0, 0);
		
	if (!$PortObj) {
		print STDERR "Can't open $PORTNAME: $!\n";
		return;
	}
	
	#$PortObj->user_msg('ON');
	$PortObj->databits(8);
	$PortObj->baudrate(19200);
	$PortObj->parity("none");
	$PortObj->stopbits(1);
	$PortObj->handshake("cts/rts");
	$PortObj->read_const_time(1000);
	
	$PortObj->write_settings || die;
	
	return $PortObj;
}

sub read_device_list {
	if (!open DEVICE_LIST, $DEVICE_LIST_FILENAME) {
		print STDERR '
Could not open your list of Insteon devices: "'.$DEVICE_LIST_FILENAME.'"
To extract the list of Insteon devices from your Indigo server, do the following:

	1. In Indigo, go to Preferences->Email and configure your outgoing mail server
	2. Go to Scripts->Open Scripts Folder 
	3. Open "example - list devices.scpt"
	4. Edit it to say "set verbose to true" instead of false
	5. Save and close
	6. Go to Window->Variable List
	7. Create new variable "noficationEmailAddress" with your email address
	8. Go to scripts and choose "example - list devices"
	9. Save the resulting email in the same directory as this script as "'.$DEVICE_LIST_FILENAME.'"
   10. Either stop the Indigo server or disconnect Indigo from the powerline interface.
   11. Run the insteon_health.pl script again
	
';	
	exit;
	}

	my %DEVICES;

	my $devicename;
	my $icounter = 0;
	
	print STDOUT "Reading device file\n";
	
	while (my $line = <DEVICE_LIST>) {
		next if ($line=~/^#/);
		
		if ($line=~/Device: (.+)\n/) {
			$devicename = $1;
			$icounter++;
			print STDOUT "\r   Reading device # $icounter";
						
		} elsif ($line=~/Address: (..\...\...)\n/) {
			$DEVICES{$1} = $devicename;
		}
	}
	close DEVICE_LIST;
	my $devices_read = keys %DEVICES;
	
	# If we didn't find any devices, then it's time to bail
	if ($devices_read == 0) {
		print STDERR "No devices found in device file $DEVICE_LIST_FILENAME";
		exit 1;
	} else {
		printf STDOUT "\nThere are %d Insteon devices to test\n", scalar(keys(%DEVICES));
		printf REPORT "There were %d Insteon devices to test\n", scalar(keys(%DEVICES));
	}
	
	return %DEVICES;
}

sub tx_command {
	my $port = shift;
	my $cmd = shift;
	
#	warn "tx_command($cmd)\n";
	
	my @cmd_bytes = split(/\s/, $cmd);
	
	foreach my $byte (@cmd_bytes) {
		$byte = pack('C', hex($byte));

		$port -> write($byte);
		select(undef, undef, undef, 0.030);
	}
}

# returns the rtt, or 0 if no response
sub rx_response {
	my ($port, $address) = @_;
	
	$port->read_char_time(0);     # don't wait for each character
	$port->read_const_time(5000); # timeout if we don't get what we're looking for

	my $buf = '';
	
	my $t_start = time;

	### accumulate one byte at a time until we see the substring we're looking for
    
	while (1) {
		my ($count_in, $string_in) = $port->read(1);

		if ($count_in == 0) {
		#	warn "TIMEOUT\n";
			return 0;
		}
		
		$buf .= $string_in;

		my $bufstring = packed_to_text($buf);
		
		# warn "bufstring: ".$bufstring;
		
		if ($bufstring =~/02 50 $address (.. .. ..) (..) (.. ..)/) {
		
			my $powerlinc_addr = $1;
			my $flags = $2;
			my $command = $3;
		
		#	warn "got response!\n";

			my $rtt = time() - $t_start;
	
			return $rtt;
		}
		
	}
}

sub poll_device {
	my ($port, $address) = @_;

	my @FLAGS = qw(00 05 0C 0F);			# see Insteon developer's guide page 43
	my $flags = $FLAGS[$INITIAL_HOPS];

	$address=~ s/\./ /g;	

	tx_command($port, "02 62 $address $flags 19 00");
	
	my $rtt = rx_response($port, $address);

	return $rtt;
}

#######################################################

sub generate_report {

	my $RTT_UNREACHABLE = 2.5;		# Fake unreachable time used for calculating average RTT
	my $GOOD_DEVICE_REACHABILITY_PERCENT = 95;
	
	my $filename = shift;
	
	open LOG, $filename || die "can't open: $filename";

	my $report = {};
	
	print STDOUT "\nGenerating report\n";

	while (my $line = <LOG>) {
		chomp ($line);
		my ($address, $ping, $devicename) = split(/\t/, $line);
		
		if (!defined($report->{$address})) {
			$report->{$address} = {
				devicename 			=> $devicename,
				measurements 	=> []
			};
		}
		
		push (@{$report->{$address}->{measurements}}, $ping);
	}
	
	#print Dumper($report);
	
	print REPORT "\n##########    Report for: $filename  ########## \n\n";
	print REPORT "Per-device statistics:                name\tsuccess\tmin\tavg\tmax\tstdev\tavg\n\n";
	
	my ($all_sum, $all_num, $all_success, $all_num_good, $num_devices) = (0, 0, 0, 0, 0);
	
	foreach my $address (sort { $report->{$a}->{devicename} cmp $report->{$b}->{devicename} } (keys(%$report))) {	
		
		my ($num, $sum, $num_success, $min, $max) = (0, 0, 0, 99999, 0);
		my ($mean, $sum_sq, $mean_sqrs, $sqr_mean, $var, $sd) = (0, 0, 0, 0, 0);
	
		$num_devices++;
		foreach my $ping (@{$report->{$address}->{measurements}}) {
			if ($ping eq '-') {
				$ping = $RTT_UNREACHABLE;
			} else {
				$num_success++;
				$all_success++;
			}
			
			$min = $ping if $ping < $min;
			$max = $ping if $ping > $max;
			
			$sum += $ping;
			$sum_sq += $ping**2;
			$num++;
			
			$all_sum += $ping;
			$all_num++;
		}
		
		# After each device (address) calculate its statistics
		$mean = $sum / $num;            # The average response time
		$mean_sqrs = $sum_sq / $num;    # The average of the sum of the response times squared
		$sqr_mean = $mean**2;           # The square of the mean    
		$var = $mean_sqrs - $sqr_mean;  # The variance
		$sd = sqrt($var);               # The standard deviation
		
		if ((100 * $num_success / $num) >= $GOOD_DEVICE_REACHABILITY_PERCENT) {
			$all_num_good++;
		}
		
		printf REPORT "%42s\t%3d\t%1.3f\t%1.3f\t%1.3f\t%1.3f\t%s\n", $report->{$address}->{devicename}, $num_success, $min, $mean, $max, $sd, ('*' x ($mean*20));
	}
	
	print REPORT "\n\nWhole-house statistics:\n\n";
	
	printf REPORT "\tSuccess rate:              %d / %d  %.1f%% \n", $all_success, $all_num, (100*$all_success/$all_num);
	printf REPORT "\tGood, bad, total devices:  %d, %d, %d            (good means >=%d%% reachable)\n", $all_num_good, $num_devices - $all_num_good, $num_devices, $GOOD_DEVICE_REACHABILITY_PERCENT;
	printf REPORT "\tAverage RTT:               %.3f\n", $all_sum / $all_num;
	printf REPORT "\n";

}

##################################################################
## to facilitate development you can re-run the statistics by giving a previous log as an argument 
if (my $report_only = shift(@ARGV)) {
	generate_report($report_only);
	exit;
}

select (STDOUT);
$| = 1;   # unbuffered STDOUT
select(REPORT);
$| = 1;

open REPORT, ">$RPTFIL" || die "can't open report file: $RPTFIL";

print REPORT "<head><title>iHealth Report</title></head>\n<body><pre>\n\n";

#print `pwd` . "\n";
my %DEVICES = read_device_list();

my $numDevices = scalar(keys(%DEVICES));

my $port = usb_connect($INTERFACE);

my $when = `date "+%Y%m%d_%H%M%S"`;
chomp $when;
$when = $LOGDIR . "/" . $when;

`mkdir -p $LOGDIR`;
open LOG, ">$when";
select((select(LOG), $|=1)[0]);

`rm $LOGDIR/latest`;
`ln -s $when $LOGDIR/latest`;

print STDOUT "\n\n";

for (my $iteration = 1; $iteration <= $ITERATIONS; $iteration++) {
	print STDOUT "\e[1A\rStarting test pass # $iteration of $ITERATIONS\n";

	print REPORT "\n\n#################################### ITERATION $iteration / $ITERATIONS ####################################\n\n";

#	print "resetting PLM please wait\n";
#	tx_command($port, "02 67");
#	sleep(20);

	sleep(1);

	my $dev_count=0;
	foreach my $address (shuffle(keys(%DEVICES))) {
		$dev_count++;
		print STDOUT "\r   Testing device # $dev_count of $numDevices";
		my $device = $DEVICES{$address};
		printf REPORT "%42s\t", $device;

		select(undef, undef, undef, 0.3);	# need a brief sleep otherwise we crash the PLM eventually
	
		my $rtt = poll_device($port, $address);
				
		if ($rtt) {
			$rtt = sprintf("%02.3f", $rtt);
			print REPORT "$rtt\t", '*' x ($rtt * 20);
		} else {
			$rtt = '-';
			print REPORT "-\t", '- - - - - - - - - - - - -';
		}

		print REPORT "\n";
		
		printf LOG "$address\t$rtt\t$device\n";
	}
}

close LOG;

generate_report("$when");

print STDOUT "Report Completed and saved to\n    $RPTFIL\n";
print REPORT "\n</pre></body></html>\n";

close REPORT;
